@@ -4,10 +4,15 @@ require 'net/http/persistent'
 class TestNetHttpPersistentTimedStackMulti < Minitest::Test
 
   class Connection
-    attr_reader :host
+    attr_reader :host, :closed
 
     def initialize(host)
       @host = host
+      @closed = false
+    end
+
+    def close
+      @closed = true
     end
   end
 
@@ -24,7 +29,7 @@ class TestNetHttpPersistentTimedStackMulti < Minitest::Test
 
     assert_empty stack
 
-    stack.push connection_args: popped
+    stack.push popped, connection_args: 'default'
 
     refute_empty stack
   end
@@ -38,7 +43,7 @@ class TestNetHttpPersistentTimedStackMulti < Minitest::Test
 
     assert_equal 0, stack.length
 
-    stack.push connection_args: popped
+    stack.push popped, connection_args: 'default'
 
     assert_equal 1, stack.length
   end
@@ -57,7 +62,7 @@ class TestNetHttpPersistentTimedStackMulti < Minitest::Test
       @stack.pop timeout: 0
     end
 
-    assert_match 'Waited 0 sec', e.message
+    assert_includes e.message, 'Waited 0 sec'
   end
 
   def test_pop_full
@@ -69,6 +74,18 @@ class TestNetHttpPersistentTimedStackMulti < Minitest::Test
     assert_empty stack
   end
 
+  def test_pop_closes_extra_connections
+    stack = Net::HTTP::Persistent::TimedStackMulti.new(1) { |host| Connection.new(host) }
+
+    a_conn = stack.pop connection_args: 'a.example'
+    stack.push a_conn, connection_args: 'a.example'
+
+    b_conn = stack.pop connection_args: 'b.example'
+
+    assert a_conn.closed
+    refute b_conn.closed
+  end
+
   def test_pop_wait
     thread = Thread.start do
       @stack.pop
@@ -96,7 +113,7 @@ class TestNetHttpPersistentTimedStackMulti < Minitest::Test
 
     conn = stack.pop
 
-    stack.push connection_args: conn
+    stack.push conn, connection_args: 'default'
 
     refute_empty stack
   end
@@ -108,14 +125,16 @@ class TestNetHttpPersistentTimedStackMulti < Minitest::Test
       called << object
     end
 
-    @stack.push connection_args: Object.new
+    obj = Object.new
+    @stack.push obj, connection_args: 'default'
 
     refute_empty called
     assert_empty @stack
   end
 
   def test_shutdown
-    @stack.push connection_args: Object.new
+    obj = Object.new
+    @stack.push obj, connection_args: 'default'
 
     called = []
 
